// TetrisGame.cpp: implementation of the CTetrisGame class.
//
//////////////////////////////////////////////////////////////////////

#include "TetrisGame.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTetrisGame::CTetrisGame()
{
    for (int i = 0; i < GAME_WIDTH; i++)
        for (int j = 0; j < GAME_HEIGHT; j++)
            m_gameState[i][j] = false;

    for (i = 0; i < GAME_WIDTH; i++)
        m_currentState[i] = 0; // height of each column

    // init game shapes

    // shape 0:    **
    //             **
    gameShapes[0].nDiffRotat = 1;
    gameShapes[0].shapeWeight = 4;
    gameShapes[0].shapeDescrip[0].gravityCenterHeight = 2;
    gameShapes[0].shapeDescrip[0].shapeDescrip[0][0] = true;
    gameShapes[0].shapeDescrip[0].shapeDescrip[0][1] = true;
    gameShapes[0].shapeDescrip[0].shapeDescrip[1][0] = true;
    gameShapes[0].shapeDescrip[0].shapeDescrip[1][1] = true;
    gameShapes[0].shapeDescrip[0].shapeHeight = 2;
    gameShapes[0].shapeDescrip[0].shapeWidth = 2;
    gameShapes[0].shapeDescrip[0].maxInColumns[0] = 2;
    gameShapes[0].shapeDescrip[0].minInColumns[0] = 0;
    gameShapes[0].shapeDescrip[0].maxInColumns[1] = 2;
    gameShapes[0].shapeDescrip[0].minInColumns[1] = 0;

    // shape 1:     *
    //             **
    gameShapes[1].nDiffRotat = 4;
    gameShapes[1].shapeWeight = 3;
    gameShapes[1].shapeDescrip[0].gravityCenterHeight = 1;
    gameShapes[1].shapeDescrip[0].shapeDescrip[0][0] = true;
    gameShapes[1].shapeDescrip[0].shapeDescrip[0][1] = false;
    gameShapes[1].shapeDescrip[0].shapeDescrip[1][0] = true;
    gameShapes[1].shapeDescrip[0].shapeDescrip[1][1] = true;
    gameShapes[1].shapeDescrip[0].shapeHeight = 2;
    gameShapes[1].shapeDescrip[0].shapeWidth = 2;
    gameShapes[1].shapeDescrip[0].maxInColumns[0] = 1;
    gameShapes[1].shapeDescrip[0].minInColumns[0] = 0;
    gameShapes[1].shapeDescrip[0].maxInColumns[1] = 2;
    gameShapes[1].shapeDescrip[0].minInColumns[1] = 0;

    gameShapes[1].shapeDescrip[1].gravityCenterHeight = 1;
    gameShapes[1].shapeDescrip[1].shapeDescrip[0][0] = true;
    gameShapes[1].shapeDescrip[1].shapeDescrip[0][1] = true;
    gameShapes[1].shapeDescrip[1].shapeDescrip[1][0] = true;
    gameShapes[1].shapeDescrip[1].shapeDescrip[1][1] = false;
    gameShapes[1].shapeDescrip[1].shapeHeight = 2;
    gameShapes[1].shapeDescrip[1].shapeWidth = 2;
    gameShapes[1].shapeDescrip[1].maxInColumns[0] = 2;
    gameShapes[1].shapeDescrip[1].minInColumns[0] = 0;
    gameShapes[1].shapeDescrip[1].maxInColumns[1] = 1;
    gameShapes[1].shapeDescrip[1].minInColumns[1] = 0;

    gameShapes[1].shapeDescrip[2].gravityCenterHeight = 2;
    gameShapes[1].shapeDescrip[2].shapeDescrip[0][0] = true;
    gameShapes[1].shapeDescrip[2].shapeDescrip[0][1] = true;
    gameShapes[1].shapeDescrip[2].shapeDescrip[1][0] = false;
    gameShapes[1].shapeDescrip[2].shapeDescrip[1][1] = true;
    gameShapes[1].shapeDescrip[2].shapeHeight = 2;
    gameShapes[1].shapeDescrip[2].shapeWidth = 2;
    gameShapes[1].shapeDescrip[2].maxInColumns[0] = 2;
    gameShapes[1].shapeDescrip[2].minInColumns[0] = 0;
    gameShapes[1].shapeDescrip[2].maxInColumns[1] = 2;
    gameShapes[1].shapeDescrip[2].minInColumns[1] = 1;

    gameShapes[1].shapeDescrip[3].gravityCenterHeight = 2;
    gameShapes[1].shapeDescrip[3].shapeDescrip[0][0] = false;
    gameShapes[1].shapeDescrip[3].shapeDescrip[0][1] = true;
    gameShapes[1].shapeDescrip[3].shapeDescrip[1][0] = true;
    gameShapes[1].shapeDescrip[3].shapeDescrip[1][1] = true;
    gameShapes[1].shapeDescrip[3].shapeHeight = 2;
    gameShapes[1].shapeDescrip[3].shapeWidth = 2;
    gameShapes[1].shapeDescrip[3].maxInColumns[0] = 2;
    gameShapes[1].shapeDescrip[3].minInColumns[0] = 1;
    gameShapes[1].shapeDescrip[3].maxInColumns[1] = 2;
    gameShapes[1].shapeDescrip[3].minInColumns[1] = 0;

    // shape 2:     *
    //             * 
    gameShapes[2].nDiffRotat = 2;
    gameShapes[2].shapeWeight = 2;
    gameShapes[2].shapeDescrip[0].gravityCenterHeight = 1;
    gameShapes[2].shapeDescrip[0].shapeDescrip[0][0] = true;
    gameShapes[2].shapeDescrip[0].shapeDescrip[0][1] = false;
    gameShapes[2].shapeDescrip[0].shapeDescrip[1][0] = false;
    gameShapes[2].shapeDescrip[0].shapeDescrip[1][1] = true;
    gameShapes[2].shapeDescrip[0].shapeHeight = 2;
    gameShapes[2].shapeDescrip[0].shapeWidth = 2;
    gameShapes[2].shapeDescrip[0].maxInColumns[0] = 1;
    gameShapes[2].shapeDescrip[0].minInColumns[0] = 0;
    gameShapes[2].shapeDescrip[0].maxInColumns[1] = 2;
    gameShapes[2].shapeDescrip[0].minInColumns[1] = 1;

    gameShapes[2].shapeDescrip[1].gravityCenterHeight = 1;
    gameShapes[2].shapeDescrip[1].shapeDescrip[0][0] = false;
    gameShapes[2].shapeDescrip[1].shapeDescrip[0][1] = true;
    gameShapes[2].shapeDescrip[1].shapeDescrip[1][0] = true;
    gameShapes[2].shapeDescrip[1].shapeDescrip[1][1] = false;
    gameShapes[2].shapeDescrip[1].shapeHeight = 2;
    gameShapes[2].shapeDescrip[1].shapeWidth = 2;
    gameShapes[2].shapeDescrip[1].maxInColumns[0] = 2;
    gameShapes[2].shapeDescrip[1].minInColumns[0] = 1;
    gameShapes[2].shapeDescrip[1].maxInColumns[1] = 1;
    gameShapes[2].shapeDescrip[1].minInColumns[1] = 0;

    // shape 3:    **
    gameShapes[3].nDiffRotat = 2;
    gameShapes[3].shapeWeight = 2;
    gameShapes[3].shapeDescrip[0].gravityCenterHeight = 0;
    gameShapes[3].shapeDescrip[0].shapeDescrip[0][0] = true;
    gameShapes[3].shapeDescrip[0].shapeDescrip[0][1] = false;
    gameShapes[3].shapeDescrip[0].shapeDescrip[1][0] = true;
    gameShapes[3].shapeDescrip[0].shapeDescrip[1][1] = false;
    gameShapes[3].shapeDescrip[0].shapeHeight = 1;
    gameShapes[3].shapeDescrip[0].shapeWidth = 2;
    gameShapes[3].shapeDescrip[0].maxInColumns[0] = 1;
    gameShapes[3].shapeDescrip[0].minInColumns[0] = 0;
    gameShapes[3].shapeDescrip[0].maxInColumns[1] = 1;
    gameShapes[3].shapeDescrip[0].minInColumns[1] = 0;

    gameShapes[3].shapeDescrip[1].gravityCenterHeight = 1;
    gameShapes[3].shapeDescrip[1].shapeDescrip[0][0] = true;
    gameShapes[3].shapeDescrip[1].shapeDescrip[0][1] = true;
    gameShapes[3].shapeDescrip[1].shapeDescrip[1][0] = false;
    gameShapes[3].shapeDescrip[1].shapeDescrip[1][1] = false;
    gameShapes[3].shapeDescrip[1].shapeHeight = 2;
    gameShapes[3].shapeDescrip[1].shapeWidth = 1;
    gameShapes[3].shapeDescrip[1].maxInColumns[0] = 2;
    gameShapes[3].shapeDescrip[1].minInColumns[0] = 0;

    // shape 4:    *
    gameShapes[4].nDiffRotat = 1;
    gameShapes[4].shapeWeight = 1;
    gameShapes[4].shapeDescrip[0].gravityCenterHeight = 0;
    gameShapes[4].shapeDescrip[0].shapeDescrip[0][0] = true;
    gameShapes[4].shapeDescrip[0].shapeDescrip[0][1] = false;
    gameShapes[4].shapeDescrip[0].shapeDescrip[1][0] = false;
    gameShapes[4].shapeDescrip[0].shapeDescrip[1][1] = false;
    gameShapes[4].shapeDescrip[0].shapeHeight = 1;
    gameShapes[4].shapeDescrip[0].shapeWidth = 1;
    gameShapes[4].shapeDescrip[0].maxInColumns[0] = 1;
    gameShapes[4].shapeDescrip[0].minInColumns[0] = 0;

    // init symetry table
    
    int symStateIdx = 0;

    for (i = 0; i < N_ORIG_STATES; i++)
    {
        int symNum = 0;

        for (int j = 0; j < GAME_WIDTH; j++)
        {
            int lowerBit = i & (1 << j);
            int upperBit = i & (1 << (j + GAME_WIDTH));

            if (lowerBit)
                symNum |= 1 << (GAME_WIDTH - j - 1);
            if (upperBit)
                symNum |= 1 << (2 * GAME_WIDTH - j - 1);
        }

        if (i == symNum)
        {
            m_symStates[i].state = symStateIdx;
            m_symStates[i].fMiror = false;
            symStateIdx++;
        }
        else if (i < symNum)
        {
            m_symStates[i].state = symStateIdx;
            m_symStates[i].fMiror = false;
            m_symStates[symNum].state = symStateIdx;
            m_symStates[symNum].fMiror = true;
            symStateIdx++;
        }
    }
    assert (symStateIdx == 2080);
}

CTetrisGame::~CTetrisGame()
{

}

int CTetrisGame::GetScore(int rot, int column)
{
    return 0;
}

int CTetrisGame::CalcLowestPlacement(int shape, int rot, int col)
{
    TetrisShape currShape = gameShapes[shape];
    int bestScore = BAD_SCORE;
    int minGrav = MAX_GRAV_SCORE;

    ShapeDescription currDescrip = currShape.shapeDescrip[rot];

    int placement = 0;
    for (int i = 0; i < currDescrip.shapeWidth; i++)
    {
        int nDown = m_currentState[col + i] - currDescrip.minInColumns[i];
        if (nDown > placement)
            placement = nDown;
    }

    return placement;
}

void CTetrisGame::GetBestAction(CTD &TD, int shape, int &rot, int &column)
{
    TetrisShape currShape = gameShapes[shape];
    float bestScore = (float)BAD_SCORE;

    // save current state of the game
	bool copyOfGameState[GAME_WIDTH][GAME_HEIGHT];
    memcpy(copyOfGameState, m_gameState, GAME_WIDTH*GAME_HEIGHT*sizeof(bool));
    int copyOfCurrentState[GAME_WIDTH];
    memcpy(copyOfCurrentState, m_currentState, GAME_WIDTH * sizeof(int));

    for (int nRot = 0; nRot < currShape.nDiffRotat; nRot++)
    {
        ShapeDescription currDescrip = currShape.shapeDescrip[nRot];

        for (int nCol = 0; nCol <= GAME_WIDTH - currDescrip.shapeWidth; nCol++)
        {
            int reward = CTetrisGame::PerformAction(shape, nRot, nCol);       

            bool topLevels[2][GAME_WIDTH];
            GetTopLevels(topLevels);
            int state = Levels2State(topLevels);

            int newState;
            bool fMiror;
            State2SymState(state, newState, fMiror);

            float val = TD.GetValue(newState);

            float score = (float)(reward + LAMBDA * val);

            if (  (score > bestScore) || 
                  ( (score == bestScore) && (rand() > 0.3*RAND_MAX) ) )
            {
                bestScore = score;
                column = nCol;
                rot = nRot;
            }

            // reconstruct previous game state
            memcpy(m_gameState, copyOfGameState, GAME_WIDTH*GAME_HEIGHT*sizeof(bool));
            memcpy(m_currentState, copyOfCurrentState, GAME_WIDTH * sizeof(int));
        }
    }
}

void CTetrisGame::GetBestAction(int *contour, int shape, int &rot, int &column, int &placement)
{
    m_currentState[0] = 2;

    for (int i = 1; i < GAME_WIDTH; i++)
        m_currentState[i] = m_currentState[i - 1] + contour[i - 1];

    GetBestAction(shape, rot, column, placement);
}

void CTetrisGame::GetBestAction(int shape, int &rot, int &column, int &placement)
{
    TetrisShape currShape = gameShapes[shape];
    int bestScore = BAD_SCORE;
    int minGrav = MAX_GRAV_SCORE;

    for (int nRot = 0; nRot < currShape.nDiffRotat; nRot++)
    {
        ShapeDescription currDescrip = currShape.shapeDescrip[nRot];
        int minHeight = GAME_HEIGHT;
        int bestColumn;

        for (int nCol = 0; nCol <= GAME_WIDTH - currDescrip.shapeWidth; nCol++)
        {
            int maxDown = 0;
            for (int i = 0; i < currDescrip.shapeWidth; i++)
            {
                int nDown = m_currentState[nCol + i] - currDescrip.minInColumns[i];
                if (nDown > maxDown)
                    maxDown = nDown;
            }

            if (maxDown < minHeight)
            {
                minHeight = maxDown;
                bestColumn = nCol;
            }
        }

        int gravScore = currShape.shapeWeight * minHeight + 
                        currDescrip.gravityCenterHeight;
        if (gravScore < minGrav)
        {
            minGrav = gravScore;
            rot = nRot;
            placement = minHeight;
            column = bestColumn;
        }
    }
}

int *CTetrisGame::GetCurrentState()
{
    return m_currentState;       
}

int CTetrisGame::DrawShape()
{
    return (rand() % N_SHAPE_TYPES);
}

void CTetrisGame::GetContour(int *contour)
{
    for (int i = 0; i < GAME_WIDTH - 1; i++)
    {
        int diff = m_currentState[i + 1] - m_currentState[i];
        if (diff < 0)
            diff = __max(-MAX_RELEVANT_HEIGHT, diff);
        else
            diff = __min(MAX_RELEVANT_HEIGHT, diff);

        contour[i] = diff;
    }
}

int CTetrisGame::PerformAction(int shape, int rot, int col)
{
    int placement = CalcLowestPlacement(shape, rot, col);
    return PerformAction(shape, rot, col, placement);
}

int CTetrisGame::PerformAction(int shape, int rot, int col, int placement)
{
    TetrisShape currShape = gameShapes[shape];
    ShapeDescription currDescrip = currShape.shapeDescrip[rot];

    // update game state and height of each column
    for (int i = 0; i < currDescrip.shapeWidth; i++)
    {
        for (int j = currDescrip.minInColumns[i]; j < currDescrip.maxInColumns[i]; j++)
        {
            if (currDescrip.shapeDescrip[i][j])
                m_gameState[col + i][placement + j] = true;
        }
        if (j != 0)
            m_currentState[col + i] = placement + j;
    }

    // remove full lines
    int reward = 0;
    for (i = placement; i < placement + currDescrip.shapeHeight; i++)
    {
        for (int j = 0; j < GAME_WIDTH; j++)
        {
            if (!m_gameState[j][i])
                break;
        }
        if (j == GAME_WIDTH) // full line - remove
        {
            for (int k = 0; k < GAME_WIDTH; k++)
            {
                int from   = (int) &(m_gameState[k][i+1]);
                int to     = (int) &(m_gameState[k][i]);
                int amount = (int) sizeof(bool) * (m_currentState[k] - i);

                assert ( (from+amount < (int) m_gameState[k+1]) && 
                         (to >= (int) m_gameState[k]) &&
                         (amount < 200) );

//                memmove(&m_gameState[k][i], &m_gameState[k][i + 1], 
//                       sizeof(bool) * (m_currentState[k] - i));
                memmove ((void *)to, (void *)from, amount);

            }
            reward++;
        }
    }

    // update the height of each column
    for (int j = 0; j < GAME_WIDTH; j++)
    {
        for (int k = m_currentState[j]; k > 0; k--)
        {
            if (m_gameState[j][k - 1])
                break;
        }
        m_currentState[j] = k;
    }

    return reward;
}

int CTetrisGame::GetHeight()
{
    for (int i = 0; i < GAME_WIDTH; i++)
    {
        printf("%d\t", m_currentState[i]);
    }
    printf("\n");
    fflush(stdout);
    return 0;
}

bool CTetrisGame::IsValidAction(int shapeIdx, int rot, int column)
{
    TetrisShape currShape = gameShapes[shapeIdx];

    if (rot >= currShape.nDiffRotat)
        return false;

    ShapeDescription currDescrip = currShape.shapeDescrip[rot];
    if (column > GAME_WIDTH - currDescrip.shapeWidth)
        return false;

    return true;
}

void CTetrisGame::GetTopLevels(bool topLevels[2][GAME_WIDTH])
{
    // find heighest column
    int max = 2;
    for (int i = 0; i< GAME_WIDTH; i++)
    {
        if (m_currentState[i] > max)
            max = m_currentState[i];
    }

    for (i = 0; i < 2; i++)
    {
        for (int j = 0; j < GAME_WIDTH; j++)
            topLevels[i][j] = m_gameState[j][max - 2 + i];
    }
}

int CTetrisGame::Levels2State(bool topLevels[2][GAME_WIDTH])
{
    int state = 0;

    for (int i = 0; i < 2; i++)
    {
        for (int j = 0; j < GAME_WIDTH; j++)
        {
            state += (topLevels[i][j] ? (int)pow(2, i * GAME_WIDTH + j) : 0);
        }
    }
    return state;
}
