// TetrisGame.h: interface for the CTetrisGame class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TETRISGAME_H__97A29B63_E6BD_11D3_B759_8780A4B95C70__INCLUDED_)
#define AFX_TETRISGAME_H__97A29B63_E6BD_11D3_B759_8780A4B95C70__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TD.h"

const int N_ORIG_STATES = 4096; // 2 ** 12
const int GAME_WIDTH = 6;
const int GAME_HEIGHT = 2000;
const int N_SHAPE_TYPES = 5;
const int MAX_SHAPE_WIDTH = 2;
const int MAX_SHAPE_HEIGHT = 2;
const int MAX_RELEVANT_HEIGHT = 2;
const int MAX_INTERVAL = MAX_RELEVANT_HEIGHT * 2 + 1;
const int BAD_SCORE = -100000;
const int MAX_GRAV_SCORE = GAME_HEIGHT * MAX_SHAPE_WIDTH * MAX_SHAPE_HEIGHT;
const int N_ROTAT = 4;

typedef struct
{
    int state;
    bool fMiror;
} SymStateType;

typedef struct
{
    int shapeWidth;
    int shapeHeight;
    int minInColumns[MAX_SHAPE_WIDTH];
    int maxInColumns[MAX_SHAPE_WIDTH];
    bool shapeDescrip[MAX_SHAPE_WIDTH][MAX_SHAPE_HEIGHT];
    int gravityCenterHeight; // relative to the middle of the lowest line in the shape
} ShapeDescription;

typedef struct
{
    int nDiffRotat;
    ShapeDescription shapeDescrip[N_ROTAT];
    int shapeWeight;
} TetrisShape;




class CTetrisGame  
{
public:
    int Levels2State(bool topLevels[2][GAME_WIDTH]);
	void GetTopLevels(bool topLevels[2][GAME_WIDTH]);
	bool IsValidAction(int shapeIdx, int rot, int column);
	int GetHeight();
	int PerformAction(int shapeIdx, int rot, int col);
	int PerformAction(int shapeIdx, int rot, int col, int placement);
	void GetContour(int *contour);
	int DrawShape();
	int * GetCurrentState();
	int GetScore(int rot, int column );
    void GetBestAction(CTD &TD, int shape, int &rot, int &column);
    void GetBestAction(int shape, int &rot, int &column, int &placement);
    void GetBestAction(int *contour, int shape, int &rot, int &column, int &placement);
	CTetrisGame();
	virtual ~CTetrisGame();
    void State2SymState(int state, int &newState, bool &fMiror) 
    { newState = m_symStates[state].state; fMiror = m_symStates[state].fMiror; }


    TetrisShape gameShapes[N_SHAPE_TYPES];

private:

    int CalcLowestPlacement(int shape, int rot, int col);

	bool m_gameState[GAME_WIDTH][GAME_HEIGHT];
    int m_currentState[GAME_WIDTH]; // height of each column
    SymStateType m_symStates[N_ORIG_STATES];
};

#endif // !defined(AFX_TETRISGAME_H__97A29B63_E6BD_11D3_B759_8780A4B95C70__INCLUDED_)
